<?php
session_start();

// Initialize or reset session counters
if (!isset($_SESSION['totalPosts'])) {
    $_SESSION['totalPosts'] = 0;
}
if (!isset($_SESSION['updatedPosts'])) {
    $_SESSION['updatedPosts'] = 0;
}

$instanceUrl = '';
$apiKey = '';
$statusMessage = '';

if ($_SERVER['REQUEST_METHOD'] === 'POST') {
    $instanceUrl = $_POST['instanceUrl'];
    $apiKey = $_POST['apiKey'];

    if (empty($instanceUrl) || empty($apiKey)) {
        $statusMessage = 'Instance URL and API key are required.';
    } else {
        $statusMessage = 'Processing...';
        // Reset session counters at the start of a new operation
        $_SESSION['totalPosts'] = 0;
        $_SESSION['updatedPosts'] = 0;
        $_SESSION['statusMessage'] = 'Processing...';
        updatePostsVisibility($instanceUrl, $apiKey);
    }
}

function updatePostsVisibility($instanceUrl, $apiKey) {
    global $statusMessage;

    $apiUrl = rtrim($instanceUrl, '/') . '/api/v1/accounts/verify_credentials';
    $headers = [
        'Authorization: Bearer ' . $apiKey,
        'Content-Type: application/json'
    ];

    // Fetch user ID
    $response = makeApiRequest($apiUrl, 'GET', $headers);
    $userId = $response['id'] ?? null;

    if (!$userId) {
        $_SESSION['statusMessage'] = 'Unable to verify credentials or fetch user ID.';
        return;
    }

    $page = 1;
    $perPage = 300;

    do {
        $url = rtrim($instanceUrl, '/') . "/api/v1/accounts/$userId/statuses?limit=$perPage&page=$page";
        $posts = makeApiRequest($url, 'GET', $headers);

        if (empty($posts)) {
            break;
        }

        foreach ($posts as $post) {
            $postId = $post['id'];
            $visibility = $post['visibility'];

            $_SESSION['totalPosts']++;

            if ($visibility !== 'followers') {
                $updateUrl = rtrim($instanceUrl, '/') . "/api/v1/statuses/$postId";
                $data = json_encode(['visibility' => 'followers']);
                makeApiRequest($updateUrl, 'PUT', $headers, $data);
                $_SESSION['updatedPosts']++;
            }
        }

        // Respect rate limiting
        sleep(300);

        $page++;
    } while (count($posts) === $perPage);

    $_SESSION['statusMessage'] = "Processed {$_SESSION['totalPosts']} posts. Updated {$_SESSION['updatedPosts']} posts to 'followers-only'.";
}

function makeApiRequest($url, $method, $headers, $data = null) {
    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL, $url);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $method);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);

    if ($data) {
        curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
    }

    $response = curl_exec($ch);
    $error = curl_error($ch);

    curl_close($ch);

    if ($error) {
        throw new Exception('Curl error: ' . $error);
    }

    return json_decode($response, true);
}
?>

<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Mastodon Post Visibility Updater</title>
    <style>
        body { font-family: Arial, sans-serif; }
        form { margin-bottom: 20px; }
        input[type="text"], input[type="submit"] { margin: 10px 0; padding: 10px; }
        #status { margin-top: 20px; }
        .counter { font-weight: bold; }
    </style>
</head>
<body>
    <h1>Mastodon Post Visibility Updater</h1>
    <form method="post" action="">
        <label for="instanceUrl">Mastodon Instance URL:</label><br>
        <input type="text" id="instanceUrl" name="instanceUrl" value="<?php echo htmlspecialchars($instanceUrl); ?>" required><br>
        <label for="apiKey">API Key:</label><br>
        <input type="text" id="apiKey" name="apiKey" value="<?php echo htmlspecialchars($apiKey); ?>" required><br>
        <input type="submit" value="Update Posts Visibility">
    </form>
    <div id="status"><?php echo htmlspecialchars($statusMessage); ?></div>
    <div class="counter">Total Posts Processed: <span id="totalPosts"><?php echo $_SESSION['totalPosts']; ?></span></div>
    <div class="counter">Posts Updated to 'Followers-Only': <span id="updatedPosts"><?php echo $_SESSION['updatedPosts']; ?></span></div>

    <script>
        function fetchStatus() {
            fetch('status.php')
                .then(response => response.json())
                .then(data => {
                    document.getElementById('totalPosts').textContent = data.totalPosts;
                    document.getElementById('updatedPosts').textContent = data.updatedPosts;
                    document.getElementById('status').textContent = data.statusMessage;
                })
                .catch(error => console.error('Error fetching status:', error));
        }

        // Update status every 5 seconds
        setInterval(fetchStatus, 5000);
        // Fetch status immediately on page load
        fetchStatus();
    </script>
</body>
</html>
